﻿/* -----------------------------------------------------------------------
EKL(ExtendKinectLibrary)ver.α 1.0.0 のサンプルです。
Jointの位置とBoneの線を描画して、簡単に求めた身長を表示します。
BoneManagerや各クラス・構造体の拡張メソッドを利用したサンプルになっています。
簡単のため、Kinectセンサーに移るプレイヤは1人であることを前提にしています。

※1.最適な実装ではありません。説明のため、非効率な実装方法を取っています。
※2.表示される身長は凡そ正しいものにはなりません。平気で10cm以上ずれます。
    (Kinectをプレイヤに対し水平にし、全身が映る様にすると実際の値に近い値を取るようになります。
     そもそも計測方法に(多分)問題があります。簡単のため今回の様な実装になっています。)
 2012/03/01 XJINE
 -----------------------------------------------------------------------*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

//追加
using Microsoft.Kinect;
using ExtendKinectLibrary;

namespace EKLSampleWithBone
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {

        KinectSensor kinect;
        BoneManager boneManager;

        SolidColorBrush JointPointBrush = new SolidColorBrush(Colors.Blue);
        SolidColorBrush BoneLineBrush = new SolidColorBrush(Colors.LightSeaGreen);
        SolidColorBrush TextBrush = new SolidColorBrush(Colors.Red);

        Ellipse JointEllipse
        {
            get
            {
                Ellipse temp = new Ellipse();
                temp.Fill = JointPointBrush;
                temp.Height = 20;
                temp.Width = 20;
                return temp;
            }
        }

        Line BoneLine
        {
            get
            {
                Line temp = new Line();
                temp.StrokeThickness = 10;
                temp.Stroke = BoneLineBrush;
                return temp;
            }
        }

        Label DataLabel
        {
            get
            {
                Label temp = new Label();
                temp.FontSize = 28;
                temp.Foreground = TextBrush;
                temp.FontWeight = FontWeights.ExtraBold;
                return temp;
            }
        }

        public MainWindow()
        {
            InitializeComponent();

            //Kinectが一台も繋がっていなかったら終了する。
            if (KinectSensor.KinectSensors.Count == 0)
            {
                MessageBox.Show("Failed to find some Kinect sensor.");
                this.Close();
            }

            //Kinectの初期化と設定
            kinect = KinectSensor.KinectSensors[0];
            kinect.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(kinect_ColorFrameReady);
            kinect.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);
            kinect.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(kinect_SkeletonFrameReady);
            kinect.SkeletonStream.Enable();
            kinect.Start();

            //引数は管理する骨格の最大数(キャパシティ)になっている。ここでは30フレーム分の骨格を管理する。
            //30Fpsで取得する場合、毎更新すれば1秒間の骨格データが保存、管理されることになる。
            boneManager = new BoneManager(30);

            //Windowイベントの設定
            this.Closed += new EventHandler(MainWindow_Closed);
            this.KeyDown += new KeyEventHandler(MainWindow_KeyDown);
        }

        /// <summary>
        /// キーボードから入力された時の処理。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void MainWindow_KeyDown(object sender, KeyEventArgs e)
        {
            switch (e.Key)
            {
                case Key.Up:
                    kinect.TiltUp();
                    return;
                case Key.Down:
                    kinect.TiltDown();
                    return;
                case Key.Left:
                    kinect.SetTiltAngle(kinect.MaxElevationAngle);
                    return;
                case Key.Right:
                    kinect.SetTiltAngle(kinect.MinElevationAngle);
                    return;
                default:
                    return;
            }
        }

        /// <summary>
        /// ウィンドウが閉じられるときの処理。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void MainWindow_Closed(object sender, EventArgs e)
        {
            //Kinectを停止してインスタンスを破棄する
            kinect.Close();
        }

        /// <summary>
        /// カメラのデータがKinectから流れて来たときの処理。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void kinect_ColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            ColorImageFrame temp = e.OpenColorImageFrame();

            if (temp != null)
            {
                //カメラ画像を表示する
                image1.Source = temp.GetBitmapSource();

                //ImageFrameを破棄する
                temp.Dispose();
            }
        }

        /// <summary>
        /// 骨格のデータがKinectから流れて来たときの処理。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void kinect_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            
            SkeletonFrame temp = e.OpenSkeletonFrame();
            
            if (temp != null)
            {
                //骨格データを貰う(ここでは一人しか映らないことを前提としている)
                Skeleton[] skeletonData = temp.SkeletonData();

                //BoneManagerの更新
                foreach (Skeleton skeleton in skeletonData)
                {
                    //追跡できている骨格があったら
                    if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                    {
                        //BoneManagerを更新する(キャパシティを超える最も古いデータは捨てられる)
                        boneManager.Update(skeleton.Joints);

                        //JointとBoneを表示する
                        ShowJointAndBone();
                        //疑似的な身長を表示する
                        ShowPlayerHeight();
                        //左腕の角度を表示する
                        ShowLeftArmAngle();
                    }
                }

                //フレームを破棄
                temp.Dispose();
            }
        }

        /// <summary>
        /// JointとBoneの位置をキャンバス上に表示する。
        /// </summary>
        void ShowJointAndBone()
        {
            //キャンバスの初期化
            canvas.Children.Clear();
            //Boneの表示
            ShowBoneLine();
            //Jointの表示(Jointが上に来る方が見栄えが良い)
            ShowJointPoint();
        }

        /// <summary>
        /// Jointの位置を表示する。
        /// </summary>
        void ShowJointPoint()
        {
            //BoneManagerの最新のフレームから、JointCollectionEを取得する(JointCollectionとほぼ同等)
            //引数に1を与えると1フレーム前、引数に5を与えると5フレーム前になり、最も古いフレーム番号はCapacity-1になる。
            foreach (Joint joint in boneManager.GetJointCollectionE(0))
            {
                //カラー画像での座標を取得する
                ColorImagePoint position = joint.Position.InColorImage(kinect);
                //Jointを示すEllipse
                Ellipse temp = JointEllipse;
                //座標を設定して
                Canvas.SetLeft(temp, position.X);
                Canvas.SetTop(temp, position.Y);
                //キャンバスに追加
                canvas.Children.Add(temp);
            }
            //Jointを表示するだけならBoneManagerを利用する必要はないが(Skeleton[]だけあれば実装可能)、
            //JointManagerがBoneManagerに含まれる(上位互換である)ことを示すために、あえてこの様な実装になっている。
        }

        /// <summary>
        /// Boneの線を表示する。
        /// </summary>
        void ShowBoneLine()
        {
            //最新のフレームに含まれるBone全てに対して処理を行う。
            foreach (Bone bone in boneManager.GetBoneCollection(0))
            {
                //骨の両端の座標を取得する
                ColorImagePoint root = bone.RootJoint.Position.InColorImage(kinect);
                ColorImagePoint end = bone.EndJoint.Position.InColorImage(kinect);
                //骨を示すLine
                Line temp = BoneLine;
                //座標を設定して
                temp.X1 = root.X;
                temp.Y1 = root.Y;
                temp.X2 = end.X;
                temp.Y2 = end.Y;
                //キャンバスに追加
                canvas.Children.Add(temp);
            }
        }

        /// <summary>
        /// 認識したプレイヤ(このアプリケーションは簡単のため1人を前提としている)の身長を表示する。
        /// </summary>
        void ShowPlayerHeight()
        {
            //ラベルを作って
            Label temp = DataLabel;
            //ラベルの表示内容を設定する
            temp.Content = "▼[" + (int)CalcHeight(boneManager.GetBoneCollection(0)) +"]";

            //頭の位置を拾ってきて
            Joint head = boneManager.GetJoint(JointType.Head, 0);
            //カラー画像での位置に変換する
            ColorImagePoint headPosition = head.Position.InColorImage(kinect);

            //ラベルを頭の横に来るように設定して
            Canvas.SetTop(temp, headPosition.Y);
            Canvas.SetLeft(temp, headPosition.X + 20);
            
            //キャンバスに追加
            canvas.Children.Add(temp);
        }

        /// <summary>
        /// 疑似的な身長を計算して返す。かなり曖昧。
        /// </summary>
        /// <param name="boneCollection"></param>
        /// <returns></returns>
        float CalcHeight(BoneCollection boneCollection)
        {
            float height = 0;
            height += boneCollection[BoneType.ShoulderC_Head].Length;
            height += boneCollection[BoneType.Spine_ShoulderC].Length;
            height += boneCollection[BoneType.HipC_Spine].Length;
            height += boneCollection[BoneType.HipC_HipL].Length;
            height += boneCollection[BoneType.HipL_KneeL].Length;
            height += boneCollection[BoneType.KneeL_AnkleL].Length;
            return height * 100 + 13;//mをcmに変換する。頭の半分のサイズだけ足す(日本人平均)。
        }

        /// <summary>
        /// 左腕の角度を表示する。
        /// </summary>
        void ShowLeftArmAngle()
        {
            //ラベルを作って
            Label temp = DataLabel;
            //ラベルの表示内容を設定する
            //「左肩から左ひじ」への骨と「左ひじから左手首」への骨の角度を取得する
            temp.Content = "▼[" + (int)(boneManager.GetBone(BoneType.ShoulderL_ElbowL,0)
                .Angle(boneManager.GetBone(BoneType.ElbowL_WristL, 0)) * 180 / Math.PI) + "]";

            //左ひじの位置を拾ってきて
            Joint elbowL = boneManager.GetJoint(JointType.ElbowLeft,0);
            //カラー画像での位置に変換する
            ColorImagePoint elbowLPosition = elbowL.Position.InColorImage(kinect);

            //ラベルを左ひじの横に来るように設定して
            Canvas.SetTop(temp, elbowLPosition.Y);
            Canvas.SetLeft(temp, elbowLPosition.X + 20);

            //キャンバスに追加
            canvas.Children.Add(temp);
        }
    
    }
}
